@@ -267,7 +267,7 @@ pub struct Arweave {
     pub name: String,
     pub units: String,
     pub base_url: Url,
-    pub crypto: Option<crypto::Provider>,
+    pub crypto: crypto::Provider,
 }
 
 impl Default for Arweave {
@@ -276,30 +276,32 @@ impl Default for Arweave {
             name: String::from("arweave"),
             units: String::from("winstons"),
             base_url: Url::from_str("https://arweave.net/").unwrap(),
-            crypto: None,
+            crypto: crypto::Provider::default(),
         }
     }
 }
 
 impl Arweave {
-    pub async fn from_keypair_path(
-        keypair_path: PathBuf,
-        base_url: Option<Url>,
-    ) -> Result<Arweave, Error> {
+    pub async fn from_keypair_path(keypair_path: PathBuf, base_url: Url) -> Result<Arweave, Error> {
         let crypto = crypto::Provider::from_keypair_path(keypair_path).await?;
-        let mut arweave = Arweave {
-            crypto: Some(crypto),
+        let arweave = Arweave {
+            base_url,
+            crypto,
             ..Default::default()
         };
-        if let Some(base_url) = base_url {
-            arweave.base_url = base_url
-        };
 
         Ok(arweave)
     }
 
-    pub fn get_crypto(&self) -> Result<&crypto::Provider, Error> {
-        self.crypto.as_ref().ok_or(Error::KeyPairNotProvided)
+    pub fn from_keypair_path_sync(keypair_path: PathBuf, base_url: Url) -> Result<Arweave, Error> {
+        let crypto = crypto::Provider::from_keypair_path_sync(keypair_path)?;
+        let arweave = Arweave {
+            base_url,
+            crypto,
+            ..Default::default()
+        };
+
+        Ok(arweave)
     }
 
     /// Returns the balance of the wallet.
@@ -310,7 +312,7 @@ impl Arweave {
         let wallet_address = if let Some(wallet_address) = wallet_address {
             wallet_address
         } else {
-            self.get_crypto()?.wallet_address()?.to_string()
+            self.crypto.wallet_address()?.to_string()
         };
         let url = self
             .base_url
@@ -355,6 +357,23 @@ impl Arweave {
         Ok(resp)
     }
 
+    pub fn process_data(&self, data: Vec<u8>) -> Result<Transaction, Error> {
+        let chunks = generate_leaves(data.clone(), &self.crypto)?;
+        let root = generate_data_root(chunks.clone(), &self.crypto)?;
+        let data_root = Base64(root.id.clone().into_iter().collect());
+        let proofs = resolve_proofs(root, None)?;
+
+        Ok(Transaction {
+            format: 2,
+            data_size: data.len() as u64,
+            data: Base64(data),
+            data_root,
+            chunks,
+            proofs,
+            ..Default::default()
+        })
+    }
+
     pub async fn create_transaction(
         &self,
         data: Vec<u8>,
@@ -363,11 +382,8 @@ impl Arweave {
         price_terms: (u64, u64),
         auto_content_tag: bool,
     ) -> Result<Transaction, Error> {
-        let chunks = generate_leaves(data.clone(), self.get_crypto()?)?;
-        let root = generate_data_root(chunks.clone(), self.get_crypto()?)?;
-        let data_root = Base64(root.id.clone().into_iter().collect());
-        let proofs = resolve_proofs(root, None)?;
-        let owner = self.get_crypto()?.keypair_modulus()?;
+        let mut transaction = self.process_data(data)?;
+        transaction.owner = self.crypto.keypair_modulus()?;
 
         let mut tags = vec![Tag::<Base64>::from_utf8_strs(
             "User-Agent",
@@ -377,7 +393,7 @@ impl Arweave {
         // Get content type from [magic numbers](https://developer.mozilla.org/en-US/docs/Web/HTTP/Basics_of_HTTP/MIME_types)
         // and include additional tags if any.
         if auto_content_tag {
-            let content_type = if let Some(kind) = infer::get(&data) {
+            let content_type = if let Some(kind) = infer::get(&transaction.data.0) {
                 kind.mime_type()
             } else {
                 "application/octet-stream"
@@ -390,6 +406,7 @@ impl Arweave {
         if let Some(other_tags) = other_tags {
             tags.extend(other_tags);
         }
+        transaction.tags = tags;
 
         // Fetch and set last_tx if not provided (primarily for testing).
         let last_tx = if let Some(last_tx) = last_tx {
@@ -400,24 +417,14 @@ impl Arweave {
             let last_tx_str = resp.text().await?;
             Base64::from_str(&last_tx_str)?
         };
+        transaction.last_tx = last_tx;
 
-        let data_len = data.len() as u64;
-        let blocks_len = data_len / BLOCK_SIZE + (data_len % BLOCK_SIZE != 0) as u64;
+        let blocks_len =
+            transaction.data_size / BLOCK_SIZE + (transaction.data_size % BLOCK_SIZE != 0) as u64;
         let reward = price_terms.0 + price_terms.1 * (blocks_len - 1);
+        transaction.reward = reward;
 
-        Ok(Transaction {
-            format: 2,
-            data_size: data_len.clone(),
-            data: Base64(data),
-            data_root,
-            tags,
-            reward,
-            owner,
-            last_tx,
-            chunks,
-            proofs,
-            ..Default::default()
-        })
+        Ok(transaction)
     }
 
     pub async fn create_transaction_from_file_path(
@@ -436,9 +443,9 @@ impl Arweave {
     /// Gets deep hash, signs and sets signature and id.
     pub fn sign_transaction(&self, mut transaction: Transaction) -> Result<Transaction, Error> {
         let deep_hash_item = transaction.to_deep_hash_item()?;
-        let deep_hash = self.get_crypto()?.deep_hash(deep_hash_item)?;
-        let signature = self.get_crypto()?.sign(&deep_hash)?;
-        let id = self.get_crypto()?.hash_sha256(&signature)?;
+        let deep_hash = self.crypto.deep_hash(deep_hash_item)?;
+        let signature = self.crypto.sign(&deep_hash)?;
+        let id = self.crypto.hash_sha256(&signature)?;
         transaction.signature = Base64(signature);
         transaction.id = Base64(id.to_vec());
         Ok(transaction)
@@ -1014,11 +1021,11 @@ impl Arweave {
     }
 
     pub fn sign_data_item(&self, mut data_item: DataItem) -> Result<DataItem, Error> {
-        data_item.owner = self.get_crypto()?.keypair_modulus()?;
+        data_item.owner = self.crypto.keypair_modulus()?;
         let deep_hash_item = data_item.to_deep_hash_item()?;
-        let deep_hash = self.get_crypto()?.deep_hash(deep_hash_item)?;
-        let signature = self.get_crypto()?.sign(&deep_hash)?;
-        let id = self.get_crypto()?.hash_sha256(&signature)?;
+        let deep_hash = self.crypto.deep_hash(deep_hash_item)?;
+        let signature = self.crypto.sign(&deep_hash)?;
+        let id = self.crypto.hash_sha256(&signature)?;
 
         data_item.signature = Base64(signature);
         data_item.id = Base64(id.to_vec());
@@ -1194,10 +1201,10 @@ impl Arweave {
                 let mut data_item = DataItem::deserialize(bytes_vec)?;
 
                 let deep_hash = self
-                    .get_crypto()?
+                    .crypto
                     .deep_hash(data_item.to_deep_hash_item()?)
                     .unwrap();
-                self.get_crypto()?
+                self.crypto
                     .verify(&data_item.signature.0, &deep_hash)
                     .unwrap();
 
@@ -1500,6 +1507,7 @@ mod tests {
     use matches::assert_matches;
     use std::{path::PathBuf, str::FromStr, time::Instant};
     use tokio::fs;
+    use url::Url;
 
     #[tokio::test]
     async fn test_cannot_post_unsigned_transaction() -> Result<(), Error> {
@@ -1507,7 +1515,7 @@ mod tests {
             PathBuf::from(
                 "tests/fixtures/arweave-key-7eV1qae4qVNqsNChg3Scdi-DpOLJPCogct4ixoq1WNg.json",
             ),
-            None,
+            Url::from_str("http://url.com").unwrap(),
         )
         .await?;
 
@@ -1536,7 +1544,7 @@ mod tests {
             PathBuf::from(
                 "tests/fixtures/arweave-key-7eV1qae4qVNqsNChg3Scdi-DpOLJPCogct4ixoq1WNg.json",
             ),
-            None,
+            Url::from_str("http://url.com").unwrap(),
         )
         .await?;
 
@@ -1582,7 +1590,7 @@ mod tests {
             PathBuf::from(
                 "tests/fixtures/arweave-key-7eV1qae4qVNqsNChg3Scdi-DpOLJPCogct4ixoq1WNg.json",
             ),
-            None,
+            Url::from_str("http://url.com").unwrap(),
         )
         .await?;
 
@@ -1656,7 +1664,7 @@ mod tests {
             PathBuf::from(
                 "tests/fixtures/arweave-key-7eV1qae4qVNqsNChg3Scdi-DpOLJPCogct4ixoq1WNg.json",
             ),
-            None,
+            Url::from_str("http://url.com").unwrap(),
         )
         .await?;
 
